unit Files;

{$MODE Delphi}

interface

uses
  SysUtils, Dialogs, Dcu1, Math, Classes, main_h, misc_fromsource;

const
  MaxLSize = 128;// RnD max
  BYTE_ORDER_LITTLE_ENDIAN = 1;
  BYTE_ORDER_BIG_ENDIAN = 0;

type
  TByteFile = file of Byte;
  PByteFile = ^TByteFile;
  Str30 = String[30];
  Str32 = String[32];
  Str4 = String[4];
  IntArr = array of Integer;
  pChunkData = ^TChunkData;
  TChunkData = record
    Name: Str4;
    Chunksize: Longword;
    Position: Longword;// position of first byte after Name + Chunksize
    FileId: Byte;// 0 == F, 1 == F2
    ForElement: Word;// for CNT3, CUS, etc. Just to let RNDT know which element it's refering to
  end;

//function DoFileClose(var f: TByteFile): Boolean;
//function DoFileReset(var f: TByteFile): Boolean;
function GetByteAtPos(const FilePos: Integer; var FT: TByteFile; const SavePos: Boolean = True): Byte;
procedure SetByteAtFile(const FilePos: Integer; var FT: TByteFile; const NewValue: Byte);
procedure Set16bitsAtFile(const FilePos: Integer; var FT: TByteFile; const NewValue: Word);
function Get16bits(const Position: Integer; var FT: TByteFile; const SavePos: Boolean = True): Word;
function GetBits(const Position: Integer; var FT: TByteFile; const BitCount: Byte): Longword;
procedure WriteHexStrToFile(const FilePos: Integer; var FT: TByteFile; const Hex: String);

function CEIdToWord(const CEId: Word): Word;
function WordToCEId(const W: Word): Word;
function GEIdToWord(const GEId: Word): Word;
function WordToGEId(const W: Word): Word;

function CUS4_TotalChunkSize(PageCount: Byte): Integer;
function WriteCount(const Count: Byte; const Value: Byte): String;
function FindElement(const El: Word; const Appearance: Integer = 1): TBytePoint;
//function SetElement(const Pos: TBytePoint; const NewEl: Word): Boolean;
function ElementCount(const El: Word; const FromX, FromY, ToX, ToY: Byte): Integer; overload;
function ElementCount(const El: Word): Integer; overload;
//procedure LevelWrite(const FromPos, ToPos: TBytePoint);
function ReadByte(var F: TByteFile): Byte;
function ReadCount(const Count: Integer; var FT: TByteFile; const Pos: Integer = -1): String;

procedure swap_numbers(var i1: integer; var i2: integer);
procedure swap_number_pairs(var x1, y1, x2, y2: integer);
function getFile16BitInteger(var f: TByteFile; byte_order: integer): Word;
procedure putFile16BitInteger(var f: TByteFile;  value: integer;  byte_order: integer);
function getFile32BitInteger(var f: TByteFile;  byte_order: integer): integer;
procedure putFile32BitInteger(var f: TByteFile;  value: integer;  byte_order: integer);
function getFileChunk(var f: TByteFile; var chunk_name: string; chunk_size: plongword;  byte_order: integer): boolean;
procedure putFileChunk(var f: TByteFile;  chunk_name: pchar;  chunk_size: integer;  byte_order: integer);
function getFileVersion(var f: TByteFile): integer;
function putFileVersion(var f: TByteFile;  version: integer): byte;
procedure ReadUnusedBytesFromFile(var f: TByteFile; bytes: byte);
procedure WriteUnusedBytesToFile(var f: TByteFile; const bytes: byte);
function getFile8Bit(var f: TByteFile): Byte;
function putFile8Bit(var f: TByteFile; const x: Byte): Byte;
function getFile16BitBE(var f: TByteFile): Word;
function getFile16BitLE(var f: TByteFile): Word;
function putFile16BitBE(var f: TByteFile;const x:Word): Byte;
function putFile16BitLE(var f: TByteFile;const x:Word): Byte;
function getFile32BitBE(var f: TByteFile): Longword;
function getFile32BitLE(var f: TByteFile): Longword;
function putFile32BitBE(var f: TByteFile; const x: Longword): Byte;
function putFile32BitLE(var f: TByteFile; const x: Longword): Byte;
function getFileChunkBE(var f: TByteFile;var s: string; const x: pLongword): Boolean;
function getFileChunkLE(var f: TByteFile;var s: string; const x: pLongword): Boolean;
procedure putFileChunkBE(var f: TByteFile;s: PChar; x: Longint);
procedure putFileChunkLE(var f: TByteFile;s: PChar; x: Longint);
function getDefaultLevelFilename(const Nr: ShortInt): String;
procedure ReadBytesFromFile(var f: TByteFile; var buffer: IntArr; bytes: integer);

const
  CUS4_Change_PageSize = 48;
  ChunkBODY = 'BODY';
  ChunkHEAD = 'HEAD';
  ChunkCUS4 = 'CUS4';

implementation

procedure ReadBytesFromFile(var f: TByteFile; var buffer: IntArr; bytes: integer);
var
  I: Integer;
  B: Byte;
begin
  for i:=0 to bytes-1 do
  if not eof(f) then
  begin
    Read(f, B);
    buffer[i]:=B;
  end;
end;

function getDefaultLevelFilename(const Nr: ShortInt): String;
begin
  if Nr = -1 then Result:='template.level'
  else Result:=IntToStrLen(Nr, 3) + '.level';
end;

function ReadByte(var f: TByteFile): Byte;
begin
  Read(f, Result);
end;

procedure WriteByte(var f: TByteFile; const B: Byte);
begin
  Write(f, B);
end;

function ReadCount(const Count: Integer; var FT: TByteFile; const Pos: Integer = -1): String;
var
  B: Byte;
  I: Integer;
begin
  if Pos <> -1 then Seek(FT, Pos);
  Result:='';
  if FileSize(FT) >= FilePos(FT)+Count then
  begin
    for I:=1 to Count do
    if not Eof(FT) then
    begin
      Read(FT, B);
      Result:=Result + Chr(B);
    end;
  end
//  else LogLinesAdd(IntToStr(FileSize(FT)) + ' (size) < ' +
//   IntToStr(FilePos(FT)) + ' (pos) + ' + IntToStr(Count) + ' (ReadCount)');
end;

procedure swap_numbers(var i1: integer; var i2: integer);
var
  help: integer;
begin
  help:=i1;
  i1:=i2;
  i2:=help;
end;

procedure swap_number_pairs(var x1, y1, x2, y2: integer);
var
help_x: integer;
help_y: integer;
begin
  help_x:=x1;
  help_y:=y1;
  x1:=x2;
  x2:=help_x;
  y1:=y2;
  y2:=help_y;
end;

function getFile16BitInteger(var f: TByteFile; byte_order: integer): Word;
var
  A, B: Byte;
begin
  if byte_order=BYTE_ORDER_BIG_ENDIAN then
  begin
    A:=ReadByte(f);
    B:=ReadByte(f);
    Result:=BytesToWord(A, B);
  end
  else
  begin
    B:=ReadByte(f);
    A:=ReadByte(f);
    Result:=BytesToWord(A, B);
  end;
end;

procedure putFile16BitInteger(var f: TByteFile;  value: integer;  byte_order: integer);
begin
  if byte_order=BYTE_ORDER_BIG_ENDIAN then
  begin
    WriteByte(f, value div 256);
    WriteByte(f, value mod 256);
  end
  else
  begin(* BYTE_ORDER_LITTLE_ENDIAN *)
    WriteByte(f, value mod 256);
    WriteByte(f, value div 256);
  end;
end;

function getFile32BitInteger(var f: TByteFile;  byte_order: integer): integer;
begin
  if byte_order=BYTE_ORDER_BIG_ENDIAN
  then
  begin
    result:= ((ReadByte(f) shl 24) or (ReadByte(f) shl 16) or (ReadByte(f) shl 8) or (ReadByte(f) shl 0));
    exit;
  end
  else
  (* BYTE_ORDER_LITTLE_ENDIAN *)
  begin
    result:= ((ReadByte(f) shl 0) or (ReadByte(f) shl 8) or (ReadByte(f) shl 16) or (ReadByte(f) shl 24));
    exit;
  end;
end;

procedure putFile32BitInteger(var f: TByteFile;  value: integer;  byte_order: integer);
begin
  if byte_order=BYTE_ORDER_BIG_ENDIAN
  then
  begin
    WriteByte(f, (value shr 24) and $ff);
    WriteByte(f, (value shr 16) and $ff);
    WriteByte(f, (value shr 8) and $ff);
    WriteByte(f, (value shr 0) and $ff);
  end
  else
  begin
    (* BYTE_ORDER_LITTLE_ENDIAN *)
    WriteByte(f, (value shr 0) and $ff);
    WriteByte(f, (value shr 8) and $ff);
    WriteByte(f, (value shr 16) and $ff);
    WriteByte(f, (value shr 24) and $ff);
  end;
end;

function getFileChunk(var f: TByteFile; var chunk_name: string; chunk_size: plongword;  byte_order: integer): boolean;
const
  chunk_name_length = 4; (* read chunk name *)
begin
  if not eof(f) then
  begin
    chunk_name:=ReadCount(chunk_name_length, f);
// ZZZ !!! fgets(chunk_name,chunk_name_length+1,f);
    if (chunk_size<>nil) then(* read chunk size *)
    begin
      chunk_size^:=getFile32BitInteger(f,byte_order);
    end;
    Result:=True;
//    if eof(f) then Result:=False// !!! or(ferror(f)
//    else Result:=True;
  end
  else Result:=False;
end;

procedure putFileChunk(var f: TByteFile;  chunk_name: pchar;  chunk_size: integer;  byte_order: integer);
var
  I: Byte;
begin
  for I:=0 to 3 do WriteByte(f, ord(chunk_name[I]));
  (* write chunk name *)
  if chunk_size>=0 then
  begin
    (* write chunk size *)
    putFile32BitInteger(f,chunk_size,byte_order);
  end;
end;

function getFileVersion(var f: TByteFile): integer;
var
version_major: integer;
version_minor: integer;
version_patch: integer;
version_build: integer;
begin
  version_major:=ReadByte(f);
  version_minor:=ReadByte(f);
  version_patch:=ReadByte(f);
  version_build:=ReadByte(f);
  begin
    result:= VERSION_IDENT(version_major,version_minor,version_patch,version_build);
    exit;
  end;
end;

function putFileVersion(var f: TByteFile;  version: integer): byte;
var
version_majorv: integer;
version_minorv: integer;
version_patchv: integer;
version_buildv: integer;
begin
  version_majorv:=VERSION_MAJOR(version);
  version_minorv:=VERSION_MINOR(version);
  version_patchv:=VERSION_PATCH(version);
  version_buildv:=VERSION_BUILD(version);
  WriteByte(f, version_majorv);
  WriteByte(f, version_minorv);
  WriteByte(f, version_patchv);
  WriteByte(f, version_buildv);
  Result:=4;
end;

procedure ReadUnusedBytesFromFile(var f: TByteFile; bytes: byte);
begin
  while (bytes > 0) and (not eof(f)) do
  begin
    dec(bytes);
    ReadByte(f);
  end;
end;

procedure WriteUnusedBytesToFile(var f: TByteFile; const bytes: byte);
var
  I: Integer;
begin
  for I:=1 to bytes do WriteByte(f, 0);
end;

function getFile8Bit(var f: TByteFile): Byte;
begin
  Result:=ReadByte(f);
end;

function putFile8Bit(var f: TByteFile; const x: Byte): Byte;
begin
  Write(f, x);
  Result:=1;
end;

function getFile16BitBE(var f: TByteFile): Word;
begin
  Result:=getFile16BitInteger(f,BYTE_ORDER_BIG_ENDIAN);
end;
function getFile16BitLE(var f: TByteFile): Word;
begin
  Result:=getFile16BitInteger(f,BYTE_ORDER_LITTLE_ENDIAN);
end;

function putFile16BitBE(var f: TByteFile;const x:Word): Byte;
begin
  putFile16BitInteger(f,x,BYTE_ORDER_BIG_ENDIAN);
  Result:=2;
end;
function putFile16BitLE(var f: TByteFile;const x:Word): Byte;
begin
  putFile16BitInteger(f,x,BYTE_ORDER_LITTLE_ENDIAN);
  Result:=2;
end;
function getFile32BitBE(var f: TByteFile): Longword;
begin
  Result:=getFile32BitInteger(f,BYTE_ORDER_BIG_ENDIAN);
end;
function getFile32BitLE(var f: TByteFile): Longword;
begin
  Result:=getFile32BitInteger(f,BYTE_ORDER_LITTLE_ENDIAN);
end;
function putFile32BitBE(var f: TByteFile; const x: Longword): Byte;
begin
  putFile32BitInteger(f,x,BYTE_ORDER_BIG_ENDIAN);
  Result:=4;
end;

function putFile32BitLE(var f: TByteFile; const x: Longword): Byte;
begin
  putFile32BitInteger(f,x,BYTE_ORDER_LITTLE_ENDIAN);
  Result:=4;
end;

function getFileChunkBE(var f: TByteFile;var s: string; const x: pLongword): Boolean;
begin
  Result:=getFileChunk(f,s,x,BYTE_ORDER_BIG_ENDIAN)
end;

function getFileChunkLE(var f: TByteFile;var s: string; const x: pLongword): Boolean;
begin
  Result:=getFileChunk(f,s,x,BYTE_ORDER_LITTLE_ENDIAN);
end;

procedure putFileChunkBE(var f: TByteFile; s: PChar; x: Longint);
begin
  putFileChunk(f,s,x,BYTE_ORDER_BIG_ENDIAN)
end;

procedure putFileChunkLE(var f: TByteFile;s: PChar; x: Longint);
begin
  putFileChunk(f,s,x,BYTE_ORDER_LITTLE_ENDIAN)
end;

{
procedure LevelWrite(const FromPos, ToPos: TBytePoint);
var
  X, Y: Byte;
  B: TBytePoint;
begin
  for Y:=FromPos.Y to ToPos.Y do
  for X:=FromPos.X to ToPos.X do
  begin
    Seek(F, FilePosForLevelPos(BP(X, Y)));
    if level.encoding_16bit_field then
    begin
      B:=WordToBytes(level.field[X, Y]);
      Write(F, B.X);
      Write(F, B.Y);
    end
    else Write(F, level.field[X, Y]);
  end;
end;

function SetElement(const Pos: TBytePoint; const NewEl: Word): Boolean;
var
  B: Byte;
begin
  Result:=False;
  if InRange(Pos.Y, 1, level.fieldy) then
  if InRange(Pos.X, 1, level.fieldx) then
  begin
    level.field[Pos.X, Pos.Y]:=NewEl;
    Seek(F, FilePosForLevelPos(Pos));
    if not level.encoding_16bit_field then
    if NewEl < 256 then
    begin
      B:=NewEl;
      Write(F, B);
      Result:=True;
    end
    else
    else
    begin
      B:=NewEl div 256;
      Write(F, B);
      B:=NewEl mod 256;
      Write(F, B);
      Result:=True;
    end;
  end;
end;
}
function ElementCount(const El: Word; const FromX, FromY, ToX, ToY: Byte): Integer; overload;
var
  I: Integer;
  X, Y: Byte;
begin
  I:=0;
  for Y:=FromY to ToY do
  for X:=FromX to ToX do
  if level.field[X, Y] = El then Inc(I);
  Result:=I;
end;

function ElementCount(const El: Word): Integer; overload;
begin
  Result:=ElementCount(El, 1, 1, level.fieldx, level.fieldy);
end;

function FindElement(const El: Word; const Appearance: Integer = 1): TBytePoint;
var
  I: Integer;
  X, Y: Byte;
begin
  Result:=BP(0, 0);
  I:=0;
  for Y:=1 to level.fieldy do
  for X:=1 to level.fieldx do
  if level.field[X, Y] = El then
  begin
    Inc(I);
    if I = Appearance then
    begin
      Result:=BP(X, Y);
      Exit;
    end;
  end;
end;

function GetBits(const Position: Integer; var FT: TByteFile; const BitCount: Byte): Longword;
var
  Bits: array [1..4] of Byte;
  I: Byte;
  OldPos: Integer;
begin
  OldPos:=FilePos(FT);
  Seek(FT, Position);
  if InRange(BitCount, 1, 4) then
   for I:=1 to BitCount do Read(FT, Bits[I]);
  case BitCount of
    1: Result:=Bits[1];
    2: Result:=Bits[1]*256+Bits[2];
    3: Result:=Bits[1]*65536+Bits[2]*256+Bits[3];
    4: Result:=Bits[1]*16777216+Bits[2]*65536+Bits[3]*256+Bits[4];
  else
    Result:=0;
  end;
  Seek(FT, OldPos);
end;

function WriteCount(const Count: Byte; const Value: Byte): String;
var
  I: Integer;
begin
  for I:=1 to Count do Result:=Result + IntToHex(Value, 2);
end;

function CEIdToWord(const CEId: Word): Word;
begin
  Result:=CEId+EL_CUSTOM_START-1;
end;

function WordToCEId(const W: Word): Word;
begin
  Result:=W+1-EL_CUSTOM_START;
end;

function GEIdToWord(const GEId: Word): Word;
begin
  Result:=GEId+EL_GROUP_START-1;
end;

function WordToGEId(const W: Word): Word;
begin
  Result:=W+1-EL_GROUP_START;
end;

function CUS4_TotalChunkSize(PageCount: Byte): Integer;
begin
  Result:=8 + 48 + 48 + (CUS4_Change_PageSize * PageCount);
end;

function Get16bits(const Position: Integer; var FT: TByteFile; const SavePos: Boolean = True): Word;
var
  B, B2: Byte;
  OP: Integer;
begin
  if SavePos then
  if FileSize(FT) >= Position then
  begin
    OP:=FilePos(FT);
    Seek(FT, Position);
    Read(FT, B);
    Read(FT, B2);
    Result:=BytesToWord(B, B2);
    Seek(FT, OP);
  end
  else Result:=0
  else
  begin
    Read(FT, B);
    Read(FT, B2);
    Result:=BytesToWord(B, B2);
  end;
end;

procedure WriteHexStrToFile(const FilePos: Integer; var ft: TByteFile; const Hex: String);
var
  OldPos: Integer;
  B: Byte;
  I: Integer;
begin
  OldPos:=System.FilePos(FT);
  if FilePos = -1 then Seek(FT, FileSize(FT))
  else Seek(FT, FilePos);
  for I:=1 to Length(Hex) div 2 do
  begin
    B:=StrToInt('$' + Copy(Hex, (I*2)-1, 2));
    Write(FT, B);
  end;
  Seek(FT, OldPos);
end;

procedure Set16bitsAtFile(const FilePos: Integer; var ft: TByteFile; const NewValue: Word);
var
  OldPos: Integer;
  BP: TBytePoint;
begin
  OldPos:=System.FilePos(FT);
  BP:=WordToBytes(NewValue);
  if FileSize(FT) >= FilePos then
  begin
    Seek(FT, FilePos);
    Write(FT, BP.X);
    Write(FT, BP.Y);
  end
  else ShowMessage('SET_BYTE_AT_POS: Out of Range (This is a bug)');
  Seek(FT, OldPos);
end;

procedure SetByteAtFile(const FilePos: Integer; var ft: TByteFile; const NewValue: Byte);
var
  OldPos: Integer;
begin
  OldPos:=System.FilePos(FT);
  if FileSize(FT) >= FilePos then
  begin
    Seek(FT, FilePos);
    Write(FT, NewValue);
  end
  else ShowMessage('SET_BYTE_AT_POS: Out of Range (This is a bug)');
  Seek(FT, OldPos);
end;

function GetByteAtPos(const FilePos: Integer; var ft: TByteFile; const SavePos: Boolean = True): Byte;
var
  OldPos: Integer;
begin
// I/O 103 bug when closing and have used RCE, read at 4680800
  if SavePos then
  if FileSize(FT) >= FilePos then
  begin
    OldPos:=System.FilePos(FT);
    Seek(FT, FilePos);
    Read(FT, Result);
    Seek(FT, OldPos);
  end
  else
  else
  begin
    Read(FT, Result);
  end;
end;

end.
