unit RNDCE1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, {$IFDEF WIN32}Windows,{$IFEND} Graphics, Dialogs, Menus,
  uPSComponent, EditBtn, StdCtrls, main_h, ExtCtrls, ComCtrls, uPSUtils, LCLIntf, IniFiles,
  uPSComponentExt, uPSComponent_Default, uPSCompiler{, uPSPreProcessor},
  SynEdit, SynHighlighterPas;

type
  { TAMain }
  TAMain = class(TForm)
    ApplicationProperties1: TApplicationProperties;
    FontDialog1: TFontDialog;
    Label1: TLabel;
    Label2: TLabel;
    LevelFile: TFileNameEdit;
    Help_Help: TMenuItem;
    Help_About: TMenuItem;
    Dialog_SaveScript: TSaveDialog;
    Level_ProcessSet: TMenuItem;
    Level_ProcessSetNoSave: TMenuItem;
    OutputMess_ErrWarn: TMenuItem;
    OutputMess_NotProcessing: TMenuItem;
    OutputMess_Always: TMenuItem;
    Options_NonScriptMess: TMenuItem;
    Options_ScriptFont: TMenuItem;
    Options_FontLog: TMenuItem;
    Options_Backup: TMenuItem;
    Options_Include: TMenuItem;
    PSImport_Classes1: TPSImport_Classes;
    Script_SaveAs: TMenuItem;
    Level_SaveAs: TMenuItem;
    Options_AutosaveLevel: TMenuItem;
    Panel1: TPanel;
    Dialog_SaveLevel: TSaveDialog;
    ScriptFile: TFileNameEdit;
    Script_Compile: TMenuItem;
    MainMenu1: TMainMenu;
    LogMemo: TMemo;
    Menu_Options: TMenuItem;
    Menu_Help: TMenuItem;
    Level_Load: TMenuItem;
    Level_Save: TMenuItem;
    Menu_Script: TMenuItem;
    Script_New: TMenuItem;
    Menu_Level: TMenuItem;
    Script_Load: TMenuItem;
    Script_Save: TMenuItem;
    Script_Exit: TMenuItem;
    Script_Run: TMenuItem;
    PS: TPSScript;
    DirDialog: TSelectDirectoryDialog;
    Splitter1: TSplitter;
    ScriptMemo: TSynEdit;
    SynPasSyn1: TSynPasSyn;
    procedure FontDialog1ApplyClicked(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure Help_AboutClick(Sender: TObject);
    procedure Help_HelpClick(Sender: TObject);
    procedure LevelFileAcceptFileName(Sender: TObject; var Value: String);
    procedure Level_LoadClick(Sender: TObject);
    procedure Level_SaveAsClick(Sender: TObject);
    procedure Level_SaveClick(Sender: TObject);
    procedure Level_ProcessSetClick(Sender: TObject);
    procedure Options_FontLogClick(Sender: TObject);
    procedure Options_ScriptFontClick(Sender: TObject);
    procedure OutputMess_ErrWarnClick(Sender: TObject);
    procedure PSAfterExecute(Sender: TPSScript);
    procedure PSCompImport(Sender: TObject; x: TPSPascalCompiler);
    procedure PSExecute(Sender: TPSScript);
    procedure PSLine(Sender: TObject);
    function PSNeedFile(Sender: TObject; const OrginFileName: string;
      var FileName, Output: string): Boolean;
    procedure ScriptFileAcceptFileName(Sender: TObject; var Value: String);
    procedure ScriptMemoChange(Sender: TObject);
    procedure Script_CompileClick(Sender: TObject);
    procedure Script_ExitClick(Sender: TObject);
    procedure Script_LoadClick(Sender: TObject);
    procedure Script_NewClick(Sender: TObject);
    procedure Script_RunClick(Sender: TObject);
    procedure Script_SaveAsClick(Sender: TObject);
    procedure Script_SaveClick(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
    procedure FixExecCheck(IsExec: Boolean);
  end; 

const
  OUTPUT_ALWAYS = 0;
  OUTPUT_NOPROCESSING = 1;
  OUTPUT_WARN_ERR = 2;

var
  AMain: TAMain;
  lvl: LevelInfo;
  CommandLineMode, Processing: Boolean;
  homedir, DefCaption: String;
  Includes: TStringList;
  modifiedlevel, modifiedscript: boolean;
  Ini: TIniFile;
  OutputMessLevel: Integer;

procedure Debug(const S: String; MessLevel: Integer = OUTPUT_ALWAYS);
function CurrentLevelNumber: Integer;

implementation

{ TAMain }

uses files_c, script1, options1, chunk_confs;

function CurrentLevelNumber: Integer;
var
  S: String;
  I: Integer;
begin
// 000.level = 0
  S:=AMain.LevelFile.FileName;
  S:=Copy(S, LastDelimiter(DirectorySeparator, S)+1, 3
  {LastDelimiter('.', S)-1});
  Result:=-1;
  if Length(S) = 0 then Exit;
  for I:=1 to Length(S) do
  if (S[I] < '0') or (S[I] > '9') then
   Exit;
  Result:=StrToInt(S);
end;

procedure Debug(const S: String; MessLevel: Integer = OUTPUT_ALWAYS);
begin
  if (MessLevel = OUTPUT_NOPROCESSING) and (Processing) then
   Inc(MessLevel);
  if OutputMessLevel >= MessLevel then
  begin
    AMain.LogMemo.Lines.Add(S);
    if CommandLineMode then
{$IFDEF UNIX}
     Writeln(S)
{$ifend}
{$IFDEF WIN32}
     Writeln(StdOut, S)
{$ifend}
  end;
end;

procedure TAMain.FixExecCheck(IsExec: Boolean);
begin
  if Processing then IsExec:=true;

  LevelFile.Enabled:=not IsExec;
  ScriptFile.Enabled:=not IsExec;
  ScriptMemo.ReadOnly:=IsExec;
  LogMemo.ReadOnly:=IsExec;
  Menu_Options.Enabled:=not IsExec;
  Menu_Level.Enabled:=not IsExec;
  Menu_Help.Enabled:=not IsExec;
  Script_New.Enabled:=not IsExec;
  Script_Load.Enabled:=not IsExec;
  Script_Save.Enabled:=not IsExec;
  Script_SaveAs.Enabled:=not IsExec;
  Script_Compile.Enabled:=not IsExec;
end;

procedure TAMain.FormResize(Sender: TObject);
begin
  LevelFile.Width:=AMain.ClientWidth - LevelFile.ButtonWidth - LevelFile.Left - 5;
  ScriptFile.Width:=AMain.ClientWidth - ScriptFile.ButtonWidth - ScriptFile.Left - 5;
end;

procedure TAMain.Help_AboutClick(Sender: TObject);
begin
  ShowMessage('RnD Scripter by Simon "Zomis" Forsberg'#13#10+
  'Contact:'#13#10'zomis2k@hotmail.com'#13#10+
  'http://www.zomis.net'#13#10'http://www.artsoft.org/forum'#13#10+
  'Compiled with Lazarus http://www.lazarus.freepascal.org'#13#10+
  'Uses PascalScript from http://www.remobjects.com');
end;

procedure TAMain.Help_HelpClick(Sender: TObject);
begin
  ShowMessage('http://www.zomis.net/rnd/rndce');
end;

procedure TAMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  Ini.WriteBool('Options', 'AutosaveLevel', Options_AutosaveLevel.Checked);
  Ini.WriteBool('Options', 'Backup', Options_Backup.Checked);
  Ini.WriteBool('Options', 'Include', Options_Include.Checked);
  Ini.WriteBool('Options', 'NonScriptMess', Options_NonScriptMess.Checked);
{
  Ini.WriteString('Options', 'LogFont', LogMemo.Font.Name);
  Ini.WriteString('Options', 'ScriptFont', ScriptMemo.Font.Name);
  Ini.WriteInteger('Options', 'LogFontSize', LogMemo.Font.Size);
  Ini.WriteInteger('Options', 'ScriptFontSize', ScriptMemo.Font.Size);
}
  Ini.Free;
  Includes.Free;
end;

procedure TAMain.FontDialog1ApplyClicked(Sender: TObject);
begin
  if (Sender as TComponent).Tag = 1 then
   ScriptMemo.Font:=FontDialog1.Font
  else LogMemo.Font:=FontDialog1.Font;
end;

procedure TAMain.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  if PS.Running then
  case MessageDlg('Script is currently running, terminate script and exit program? (neither script or level will be saved)', mtError, mbOKCancel, 0) of
    mrOK:
    begin
      PS.Stop;
      Exit;
    end;
    mrCancel:
    begin
      CanClose:=False;
      Exit;
    end;
  end;
  if modifiedlevel then
  case MessageDlg('The level ' + LevelFile.FileName + ' might have been changed.'#13#13'Do you want to save the changes?', mtConfirmation, mbYesNoCancel, 0) of
    mrYes: Level_SaveClick(Sender);
    mrNo: ;
    mrCancel:
    begin
      CanClose:=False;
      Exit;
    end;
  end;
  if modifiedscript then
  case MessageDlg('The script ' + ScriptFile.FileName + ' has been changed.'#13#13'Do you want to save the changes?', mtConfirmation, mbYesNoCancel, 0) of
    mrYes: Script_SaveClick(Sender);
    mrNo: ;
    mrCancel:
    begin
      CanClose:=False;
      Exit;
    end;
  end;
end;

procedure TAMain.FormCreate(Sender: TObject);
var
  I: Integer;
{$IFDEF WIN32}
  PIDL : PItemIDList;
  Folder : array[0..MAX_PATH] of Char;
const
  CSIDL_PERSONAL = $0005;
{$ifend}
begin
  Includes:=TStringList.Create;
{$IFDEF UNIX}
// since LevelFile can't detect hidden directories, find hidden directory directly
  homedir:=ExpandFileName('~/') + '.rocksndiamonds';
{$ifend}
{$IFDEF WIN32}
  SHGetSpecialFolderLocation(0, CSIDL_PERSONAL, PIDL);
  SHGetPathFromIDList(PIDL, Folder);
  homedir:=Folder + '\Rocks''n''Diamonds';
{$ifend}
  if DirectoryExists(homedir) then
  begin
    LevelFile.InitialDir:=homedir + DirectorySeparator + 'levels';
    DirDialog.InitialDir:=LevelFile.InitialDir;
    Dialog_SaveLevel.InitialDir:=LevelFile.InitialDir;
    Dialog_SaveScript.InitialDir:=homedir;
    ScriptFile.InitialDir:=homedir;
    Ini:=TIniFile.Create(homedir + DirectorySeparator + 'rndscripter.ini');
  end
  else
  begin
    Ini:=TIniFile.Create('rndscripter.ini');
    Debug(homedir + ' directory not found.');
    Debug('Inifile created at working directory');
  end;
  Debug('Inifile location: ' + Ini.FileName);
  Options_AutosaveLevel.Checked:=Ini.ReadBool('Options', 'AutosaveLevel', Options_AutosaveLevel.Checked);
  Options_Backup.Checked:=Ini.ReadBool('Options', 'Backup', Options_Backup.Checked);
  Options_Include.Checked:=Ini.ReadBool('Options', 'Include', Options_Include.Checked);
  OutputMessLevel:=Ini.ReadInteger('Options', 'OutputMess', OutputMess_Always.Tag);
  for I:=0 to Options_NonScriptMess.Count-1 do
   Options_NonScriptMess.Items[I].Checked:=(I = OutputMessLevel);

  setLevelInfoToDefaults(@lvl, elements1);
{
s$IFDEF UNIXs
  LogMemo.Font.Name:=Ini.ReadString('Options', 'LogFont', '-misc-fixed-medium-r-semicondensed-*-*-120-*-*-*-*-koi8-r');
  ScriptMemo.Font.Name:=Ini.ReadString('Options', 'ScriptFont', '-adobe-courier-medium-r-normal-*-*-140-*-*-m-*-iso8859-1');
s$ifends
s$IFDEF WIN32s
  LogMemo.Font.Name:=Ini.ReadString('Options', 'LogFont', 'fixedsys');
  ScriptMemo.Font.Name:=Ini.ReadString('Options', 'ScriptFont', 'Courier');
s$ifends
  LogMemo.Font.Size:=Ini.ReadInteger('Options', 'LogFontSize', -12);
  ScriptMemo.Font.Size:=Ini.ReadInteger('Options', 'ScriptFontSize', -16);
}
  modifiedlevel:=false;
  modifiedscript:=false;
end;

function TAMain.PSNeedFile(Sender: TObject; const OrginFileName: string;
  var FileName, Output: string): Boolean;
var
  S: TStringList;
  F: String;
begin
  if FileExists(ExtractFilePath(ScriptFile.FileName) + FileName) then
   F:=ExtractFilePath(ScriptFile.FileName) + FileName
  else if FileExists(ExtractFilePath(Application.ExeName) + Filename) then
   F:=ExtractFilePath(Application.ExeName) + Filename
  else if FileExists(homedir + DirectorySeparator + FileName) then
   F:=homedir + DirectorySeparator + FileName;
  if FileExists(F) then
  begin
    S:=TStringList.Create;
    S.LoadFromFile(F);
    Output:=S.Text;
    S.Free;
    Result:=True;
    Debug('Including file "' + F + '"', OUTPUT_NOPROCESSING);
  end
  else
  begin
    Output:='';
// for each public release of rndscripter.zce, {$DEFINE defscript<x>}

{
get function/variable/constant list:
get token before . (for example: level.)
find variable - get type of token (LevelInfo)
find type - check structure members (.name, .author, .field...)
create list of structure members

if no . then search through PSscript to get all names - lowercase them
}
//  1.0 save font of log and script
//? 1.1 input choice
//  1.2 Ctrl+Space to get function/variable/constant list
//? 1.3 Inputfile, textfile support
//? setlevelinfotodefaults
//#BETA 2 change chararrs to string
//#BETA 2 add support for GEs
//#BETA 2 fix problems with default description of CE names, and fix SaveLevel_NOTE
//#BETA x warn when opening new level or script - or exiting without saving (option for script, always for level?)
//#BETA x use homedir and tinifiles for options (registry is not compatible with linux, easier to export inifiles)
//#BETA 3 loadlevel(integer) - use other element_info and level. loads 00x.level file
//#BETA 3 setelementinfotodefaults/clearelement
//#BETA 3 fix InitElementPropertiesStatic;
//#BETA 3 inputint, str, bool
//#BETA 3 fix rndscripter.zce - include as document. (update to latest version)
//#BETA 3 do scripts on whole levelset
//#BETA 3 commandline options
//#BETA 3 syntax highlightning
    Debug('Script "' + FileName + '" was not found, trying to compile without it', OUTPUT_NOPROCESSING);
    Result:=True;
  end;
end;

{procedure TAMain.PSProcessDirective(Sender: TPSPreProcessor;
  Parser: TPSPascalPreProcessorParser; const Active: Boolean;
  const DirectiveName, DirectiveParam: String; var Continue: Boolean);
begin
// use this somehow? like LEVEL ... or $RELOAD, $SETVAR...
end;

procedure TAMain.PSProcessUnknowDirective(Sender: TPSPreProcessor;
  Parser: TPSPascalPreProcessorParser; const Active: Boolean;
  const DirectiveName, DirectiveParam: String; var Continue: Boolean);
begin
// use this somehow? like $LEVEL ... or $RELOAD, $SETVAR...
end;
}
procedure TAMain.ScriptFileAcceptFileName(Sender: TObject; var Value: String);
begin
  if modifiedscript then
  case MessageDlg('The script ' + ScriptFile.FileName + ' has been changed.'#13#13'Do you want to save the changes?', mtConfirmation, mbYesNoCancel, 0) of
    mrYes: Script_SaveClick(Sender);
    mrNo:;
    mrCancel:
    begin
      Value:=ScriptFile.FileName;
      Exit;
    end;
  end;
  ScriptMemo.Lines.LoadFromFile(Value);
end;

procedure TAMain.ScriptMemoChange(Sender: TObject);
begin
  modifiedscript:=True;
end;

procedure TAMain.Script_CompileClick(Sender: TObject);
var
  I: Integer;
  S: String;
begin
  PS.Script.Clear;
  if Options_Include.Checked and
   (ScriptMemo.Lines.IndexOf('{$I rndscripter.zce}') = -1) then
    ScriptMemo.Lines.Insert(0, '{$I rndscripter.zce}');
  PS.Script.Text:=ScriptMemo.Lines.Text;
  if PS.Compile then
  begin
    S:='';
    for I:=0 to PS.CompilerMessageCount - 1 do
     S:=S + #13#10 + PS.CompilerMessages[I].MessageToString;
    if S <> '' then ShowMessage('Compile successful'+#13#10+S)
    else ShowMessage('Compile successful');
  end
  else
  begin
    S:='';
    for I:=0 to PS.CompilerMessageCount - 1 do
     S:=S + #13#10 + PS.CompilerMessages[I].MessageToString;
    ShowMessage(S);
  end;
end;

procedure TAMain.Script_ExitClick(Sender: TObject);
begin
  Close;
end;

procedure TAMain.Script_LoadClick(Sender: TObject);
begin
//  ScriptFile.Execute;
  ShowMessage('Click the button next to the editbox "script file"');
end;

procedure TAMain.Script_NewClick(Sender: TObject);
begin
  ScriptMemo.Lines.Clear;
end;

procedure TAMain.Script_RunClick(Sender: TObject);
var
  I: Integer;
  S: String;
begin
  if PS.Running then
  begin// halt execution if script is already running
    PS.Stop;
    Processing:=False;
    ShowMessageScr('Execution halted by user');
    Exit;// do not run script again, make time to edit script
  end;
  PS.Script.Clear;
  if Options_Include.Checked and
   (ScriptMemo.Lines.IndexOf('{$I rndscripter.zce}') = -1) then
    ScriptMemo.Lines.Insert(0, '{$I rndscripter.zce}');
  PS.Script.Text:=ScriptMemo.Lines.Text;
  if PS.Compile then
  begin
    PS.SetPointerToData('level', @lvl, PS.FindNamedType('LevelInfo'));
    PS.SetPointerToData('element_info', @elements1, PS.FindNamedType('ElemInfoArr'));
    if not Processing then
    begin
      DefCaption:=Caption;
      Caption:=DefCaption + ' - RUNNING SCRIPT';
    end;
    if PS.Execute then
    begin
      Debug('Script Executed', OUTPUT_NOPROCESSING);
      modifiedlevel:=True;
      if Options_AutosaveLevel.Checked and (not CommandLineMode)
       and (not Processing) then
        Level_SaveClick(Sender);
    end
    else ShowMessageScr(PS.ExecErrorToString);
    if not Processing then Caption:=DefCaption;
  end
  else
  begin
    S:='';
    for I:=0 to PS.CompilerMessageCount - 1 do
     S:=S + #13#10 + PS.CompilerMessages[I].MessageToString;
    ShowMessageScr(S);
  end;
end;

procedure TAMain.Script_SaveAsClick(Sender: TObject);
begin
  if Dialog_SaveScript.Execute then
  begin
    ScriptFile.FileName:=Dialog_SaveScript.FileName;
    Script_SaveClick(Sender);
  end;
end;

procedure TAMain.Script_SaveClick(Sender: TObject);
begin
  if ScriptFile.FileName = '' then Script_SaveAsClick(Sender)
  else if (not FileExists(ScriptFile.FileName)) or
   (MessageDlg('This will overwrite ' + ScriptFile.FileName + ', continue?', mtWarning, mbYesNo, 0) = mrYes) then
  begin
    ScriptMemo.Lines.SaveToFile(ScriptFile.FileName);
    modifiedscript:=False;
  end;
end;

procedure TAMain.LevelFileAcceptFileName(Sender: TObject; var Value: String);
begin
  if modifiedlevel then
  case MessageDlg('The level ' + LevelFile.FileName + ' might have been changed.'#13#13'Do you want to save the changes?', mtConfirmation, mbYesNoCancel, 0) of
    mrYes: Level_SaveClick(Sender);
    mrNo:;
    mrCancel:
    begin
      Value:=LevelFile.FileName;
      Exit;
    end;
  end;
  if not Processing then LogMemo.Lines.Clear;
  setLevelInfoToDefaults(@lvl, elements1);
  LoadLevelFromFileInfo_RND(@lvl, elements1, Value);
  Debug('Loaded level, ' + Chunknames, OUTPUT_NOPROCESSING);
end;

procedure TAMain.Level_LoadClick(Sender: TObject);
begin
  ShowMessage('Click the button next to the editbox "level file"');
end;

procedure TAMain.Level_SaveAsClick(Sender: TObject);
begin
  if Dialog_SaveLevel.Execute then
  begin
    LevelFile.FileName:=Dialog_SaveLevel.FileName;
    Level_SaveClick(Sender);
  end;
end;

procedure TAMain.Level_SaveClick(Sender: TObject);
begin
  if CommandLineMode or (Sender = Level_ProcessSet) then
  begin
// always backup in commandline mode?
    if (not RenameFile(LevelFile.FileName, LevelFile.FileName + '.bak')) then
     Debug('Backup file could not be created');
    SaveLevelFromFilename(@lvl, PChar(LevelFile.FileName))
  end
  else
  begin
    if LevelFile.FileName = '' then Level_SaveAsClick(Sender)
    else if Options_Backup.Checked and (not RenameFile(LevelFile.FileName, LevelFile.FileName + '.bak')) then
     if MessageDlg('Could not create backup of old level, overwrite current level anyway?', mtWarning, mbYesNo, 0) = mrYes then
      SaveLevelFromFilename(@lvl, PChar(LevelFile.FileName))
     else
    else SaveLevelFromFilename(@lvl, PChar(LevelFile.FileName));
  end;
end;

procedure TAMain.Level_ProcessSetClick(Sender: TObject);
var
  I: Integer;
  S, D: String;
  SkipSave: Boolean;
begin
// load directory, process all xxx.level files
  DefCaption:=Caption;
  SkipSave:=(Sender = Level_ProcessSetNoSave);
  LogMemo.Lines.Clear;
  if PS.Compile then
  if DirDialog.Execute then
  begin
    Processing:=True;
    D:=ExtractFilePath(DirDialog.FileName{$IFDEF WIN32}+ DirectorySeparator{$IFEND});
    Debug('Processing Levelset: ' + D);
    for I:=0 to 999 do
    if Processing then// processing can be stopped by user interaction (stopping script execution)
    begin
      S:=D + IntToStrLen(I, 3) + '.level';
      if FileExists(S) then
      begin
// load level, call script - halt on execerror, save level
        Caption:=DefCaption + ' - Processing ' + IntToStrLen(I, 3) + '.level';
        LevelFile.FileName:=S;
        Debug('Processing Level: ' + S, OUTPUT_NOPROCESSING);
        modifiedlevel:=false;
        LevelFileAcceptFileName(Sender, S);
        Script_RunClick(Sender);
        if not SkipSave then Level_SaveClick(Level_ProcessSet);
      end;
    end;
    modifiedlevel:=false;// don't show "level might have been changed"
  end
  else
  else ShowMessage('Compiling Script failed');
  Processing:=False;
  Caption:=DefCaption;
  FixExecCheck(False);
end;

procedure TAMain.Options_FontLogClick(Sender: TObject);
begin
  FontDialog1.Font:=LogMemo.Font;
  if FontDialog1.Execute then LogMemo.Font:=FontDialog1.Font;
end;

procedure TAMain.Options_ScriptFontClick(Sender: TObject);
begin
  FontDialog1.Tag:=(Sender as TComponent).Tag;
  FontDialog1.Font:=ScriptMemo.Font;
  if FontDialog1.Execute then ScriptMemo.Font:=FontDialog1.Font;
//  ShowMessage(FontDialog1.Font.Name);
//***ZZZ*** fix fontname, save and load
//  FontNameToPangoFontDescStr();
end;

procedure TAMain.OutputMess_ErrWarnClick(Sender: TObject);
var
  I: Integer;
begin
  if not (Sender is TMenuItem) then Exit;
  OutputMessLevel:=(Sender as TMenuItem).Tag;
  for I:=0 to Options_NonScriptMess.Count-1 do
   Options_NonScriptMess.Items[I].Checked:=
   (Options_NonScriptMess.Items[I].Tag = OutputMessLevel);
  Ini.WriteInteger('Options', 'OutputMess', (Sender as TMenuItem).Tag);
end;

procedure TAMain.PSAfterExecute(Sender: TPSScript);
begin
  FixExecCheck(False);// reenable options
end;

procedure TAMain.PSCompImport(Sender: TObject; x: TPSPascalCompiler);
begin
  Script_OnUses;
end;

procedure TAMain.PSExecute(Sender: TPSScript);
begin
  FixExecCheck(True);// disable some options to prevent corruption
end;

procedure TAMain.PSLine(Sender: TObject);
begin
  Application.ProcessMessages;
  // enable form responding - halt execution of script by pressing Run again
end;

initialization
  {$I rndce1.lrs}

end.

