unit rndtypes;

{$MODE Delphi}

interface

uses Files, Dialogs, SysUtils, main_h, RNDCE1;

type
  TFILE = TByteFile;
  pFILE = PByteFile;
  TIntArr = array [0..Pred(MAX_ELEMENTS_IN_GROUP)] of integer;
  PIntArr = ^TIntArr;
  CharArr = array [0..Pred(MAX_ENVELOPE_TEXT_LEN)] of Char;
  PCharArr = ^CharArr;

const
  SAVE_CONF_NEVER = 0;
  SAVE_CONF_ALWAYS = 1;
  SAVE_CONF_WHEN_CHANGED = -1;

(* values for chunks using micro chunks *)
  CONF_MASK_1_BYTE = $00;
  CONF_MASK_2_BYTE = $40;
  CONF_MASK_4_BYTE = $80;
  CONF_MASK_MULTI_BYTES = $c0;
  CONF_MASK_BYTES = $c0;
  CONF_MASK_TOKEN = $3f;

const
  TYPE_BOOLEAN = (1 shl 0);
  TYPE_INTEGER = (1 shl 6);
  TYPE_STRING = (1 shl 7);
  TYPE_BITFIELD = (1 shl 16);
  TYPE_ELEMENT = (1 shl 17);
  TYPE_CONTENT = (1 shl 18);
  TYPE_ELEMENT_LIST = (1 shl 19);
  TYPE_CONTENT_LIST = (1 shl 20);

function CONF_VALUE_1_BYTE(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_2_BYTE(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_4_BYTE(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_MULTI_BYTES(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_8_BIT(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_16_BIT(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_32_BIT(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_BYTES(x: integer): integer; {<= !!!6 unknown macro}
function CONF_VALUE_NUM_BYTES(x: integer): integer; {<= !!!6 unknown macro}

const CONF_CONTENT_NUM_ELEMENTS = (3*3);
const CONF_CONTENT_NUM_BYTES = (CONF_CONTENT_NUM_ELEMENTS*2);
const CONF_ELEMENT_NUM_BYTES = (2);

function CONF_ENTITY_NUM_BYTES(t: integer): integer; {<= !!!6 unknown macro}
function CONF_ELEMENT_BYTE_POS(i: integer): integer; {<= !!!6 unknown macro}
function CONF_ELEMENTS_ELEMENT(b: array of integer; i: integer): integer;
function CONF_CONTENT_ELEMENT_POS(c: integer; x: integer; y: integer): integer; {<= !!!6 unknown macro}
function CONF_CONTENT_BYTE_POS(c: integer; x: integer; y: integer): integer; {<= !!!6 unknown macro}
function CONF_CONTENTS_ELEMENT(b: array of integer; c: integer; x: integer; y: integer): integer; {<= !!!6 unknown macro}

(* temporary variables used to store pointers to structure members *)
type
LevelFileConfigInfo = record
element: integer; (* element for which data is to be stored *)
save_type: integer; (* save data always, never or when changed *)
data_type: integer; (* data type (used internally, not stored) *)
conf_type: integer; (* micro chunk identifier (stored in file) *)

(* (mandatory) *)
value: pointer; (* variable that holds the data to be stored *)
default_value: int64; (* initial default value for this variable *)

(* (optional) *)
value_copy: pointer; (* variable that holds the data to be copied *)
num_entities: pinteger; (* number of entities for multi-byte data *)
default_num_entities: integer; (* default number of entities for this data *)
max_num_entities: integer; (* maximal number of entities for this data *)
default_string: PString; (* optional default string for string data *)
end;
pLevelFileConfigInfo = ^LevelFileConfigInfo;

function LoadLevel_MicroChunk(var filea: TByteFile; const conf: array of LevelFileConfigInfo;  element: integer;  real_element: integer): integer;
function SetLevel_MicroChunk(hexdata: string; const conf: array of LevelFileConfigInfo;  element: integer;  real_element: integer): string;
procedure setConfigToDefaultsFromConfigList(conf: array of LevelFileConfigInfo);
function SaveLevel_MicroChunk(var filea: TByteFile; entry: pLevelFileConfigInfo; write_element: boolean): integer;
//function SaveLevel_MicroChunk(const filea: TByteFile; entry: pLevelFileConfigInfo; write_element: boolean): integer;

implementation

{uses RNDDB1;}

function getMappedElement(element: integer): integer;
begin(* remap some (historic, now obsolete) elements *)
  case element of
    EL_PLAYER_OBSOLETE:
    begin
      element:= EL_PLAYER_1;
    end;
    EL_KEY_OBSOLETE:
    begin
//      element:= EL_KEY_1;
      element:= EL_EM_KEY_1;
      {goto next_label;}{<= !!!d case label without "break"}
    end;
    EL_EM_KEY_1_FILE_OBSOLETE:
    begin
      element:= EL_EM_KEY_1;
    end;
    EL_EM_KEY_2_FILE_OBSOLETE:
    begin
      element:= EL_EM_KEY_2;
    end;
    EL_EM_KEY_3_FILE_OBSOLETE:
    begin
      element:= EL_EM_KEY_3;
    end;
    EL_EM_KEY_4_FILE_OBSOLETE:
    begin
      element:= EL_EM_KEY_4;
    end;
    EL_ENVELOPE_OBSOLETE:
    begin
      element:= EL_ENVELOPE_1;
    end;
    EL_SP_EMPTY:
    begin
      element:= EL_EMPTY;
    end;
    else

    begin
      if element>=NUM_FILE_ELEMENTS
      then
      begin
//        Error(ERR_WARN,'invalid level element %d',element);
        element:= EL_UNKNOWN;
      end;
    end;
  end;{case?}

  result:=element;
end;

{CONF_VALUE_1_BYTE(x)  (CONF_MASK_1_BYTE | (x))}
function CONF_VALUE_1_BYTE(x: integer): integer;
begin
  result:= (CONF_MASK_1_BYTE or (x))
end;

{CONF_VALUE_2_BYTE(x)  (CONF_MASK_2_BYTE | (x))}
function CONF_VALUE_2_BYTE(x: integer): integer;
begin
  result:= (CONF_MASK_2_BYTE or (x))
end;

{CONF_VALUE_4_BYTE(x)  (CONF_MASK_4_BYTE | (x))}
function CONF_VALUE_4_BYTE(x: integer): integer;
begin
  result:= (CONF_MASK_4_BYTE or (x))
end;

{CONF_VALUE_MULTI_BYTES(x) (CONF_MASK_MULTI_BYTES | (x))}
function CONF_VALUE_MULTI_BYTES(x: integer): integer;
begin
  result:= (CONF_MASK_MULTI_BYTES or (x))
end;

{CONF_VALUE_8_BIT(x)  CONF_VALUE_1_BYTE(x)}
function CONF_VALUE_8_BIT(x: integer): integer;
begin
  result:= CONF_VALUE_1_BYTE(x)
end;

{CONF_VALUE_16_BIT(x)  CONF_VALUE_2_BYTE(x)}
function CONF_VALUE_16_BIT(x: integer): integer;
begin
  result:= CONF_VALUE_2_BYTE(x)
end;

{CONF_VALUE_32_BIT(x)  CONF_VALUE_4_BYTE(x)}
function CONF_VALUE_32_BIT(x: integer): integer;
begin
  result:= CONF_VALUE_4_BYTE(x)
end;

{CONF_VALUE_BYTES(x)  CONF_VALUE_MULTI_BYTES(x)}
function CONF_VALUE_BYTES(x: integer): integer;
begin
  result:= CONF_VALUE_MULTI_BYTES(x)
end;

{CONF_VALUE_NUM_BYTES(x)  ((x) == CONF_MASK_1_BYTE ? 1 :       (x) == CONF_MASK_2_BYTE ? 2 :       (x) == CONF_MASK_4_BYTE ? 4 : 0)}
function CONF_VALUE_NUM_BYTES(x: integer): integer;
begin
  if (x)=CONF_MASK_1_BYTE then result:=1
  else if (x)=CONF_MASK_2_BYTE then result:=2
  else if (x)=CONF_MASK_4_BYTE then result:=4
  else result:=0;
end;

{CONF_ENTITY_NUM_BYTES(t) ((t) == TYPE_ELEMENT ||        (t) == TYPE_ELEMENT_LIST ?       CONF_ELEMENT_NUM_BYTES :       (t) == TYPE_CONTENT ||        (t) == TYPE_CONTENT_LIST ?       CONF_CONTENT_NUM_BYTES : 1)}
function CONF_ENTITY_NUM_BYTES(t: integer): integer;
begin
  if (t = TYPE_ELEMENT) or (t = TYPE_ELEMENT_LIST) then result:=CONF_ELEMENT_NUM_BYTES
  else if (t = TYPE_CONTENT) or (t = TYPE_CONTENT_LIST) then result:=CONF_CONTENT_NUM_BYTES
  else result:=1;
end;

{CONF_ELEMENT_BYTE_POS(i) ((i) * CONF_ELEMENT_NUM_BYTES)}
function CONF_ELEMENT_BYTE_POS(i: integer): integer;
begin
  result:= ((i)*CONF_ELEMENT_NUM_BYTES)
end;

{CONF_ELEMENTS_ELEMENT(b,i)     ((b[CONF_ELEMENT_BYTE_POS(i)] << 8) |       (b[CONF_ELEMENT_BYTE_POS(i) + 1]))}
function CONF_ELEMENTS_ELEMENT(b: array of integer; i: integer): integer;
begin
  result:= ((b[CONF_ELEMENT_BYTE_POS(i)] shl 8) or (b[CONF_ELEMENT_BYTE_POS(i)+1]))
end;

{CONF_CONTENT_ELEMENT_POS(c,x,y) ((c) * CONF_CONTENT_NUM_ELEMENTS +          (y) * 3 + (x))}
function CONF_CONTENT_ELEMENT_POS(c: integer; x: integer; y: integer): integer;
begin
  result:= ((c)*CONF_CONTENT_NUM_ELEMENTS+(y)*3+(x))
end;

{CONF_CONTENT_BYTE_POS(c,x,y) (CONF_CONTENT_ELEMENT_POS(c,x,y) *          CONF_ELEMENT_NUM_BYTES)}
function CONF_CONTENT_BYTE_POS(c: integer; x: integer; y: integer): integer;
begin
  result:= (CONF_CONTENT_ELEMENT_POS(c,x,y)*CONF_ELEMENT_NUM_BYTES)
end;

{CONF_CONTENTS_ELEMENT(b,c,x,y) ((b[CONF_CONTENT_BYTE_POS(c,x,y)]<< 8)|     (b[CONF_CONTENT_BYTE_POS(c,x,y) + 1]))}
function CONF_CONTENTS_ELEMENT(b: array of integer; c: integer; x: integer; y: integer): integer;
begin
  result:= ((b[CONF_CONTENT_BYTE_POS(c,x,y)] shl 8) or (b[CONF_CONTENT_BYTE_POS(c,x,y)+1]))
end;

procedure setConfigToDefaultsFromConfigList(conf: array of LevelFileConfigInfo);
var
i: integer;
default_value: integer;
data_type: integer;
conf_type: integer;
byte_mask: integer;
default_num_entities: integer;
max_num_entities: integer;
default_string: PString;
element_array: array of integer;
content: PContentArr;
j, c, x, y: integer;
begin
  i:=0;
  while conf[i].data_type<>-1 do
  begin
    default_value:=conf[i].default_value;
    data_type:=conf[i].data_type;
    conf_type:=conf[i].conf_type;
    byte_mask:=conf_type and CONF_MASK_BYTES;
    if byte_mask=CONF_MASK_MULTI_BYTES then
    begin
      default_num_entities:=conf[i].default_num_entities;
      max_num_entities:=conf[i].max_num_entities;
      pinteger(conf[i].num_entities)^:=default_num_entities;

      if data_type=TYPE_STRING then
      begin
        default_string:=PString(conf[i].default_string);
        PString(conf[i].value)^:=default_string^;
      end
      else
      if data_type=TYPE_ELEMENT_LIST then
      begin
        element_array:={!!!a type cast? =>} (conf[i].value);
        for j:=0 to Pred(max_num_entities) do
          element_array[j]:= default_value;
      end
      else
      if data_type=TYPE_CONTENT_LIST then
      begin
        content:=PContentArr(conf[i].value);
        for c:=0 to Pred(max_num_entities) do
        for y:=0 to Pred(3) do
        for x:=0 to Pred(3) do
          content[c].e[x, y]:= default_value;
      end;
    end
    else
    begin
      (* constant size configuration data (1, 2 or 4 bytes) *)
      if data_type=TYPE_BOOLEAN then
       pboolean(conf[i].value)^:=longbool(default_value)
      else
      pinteger(conf[i].value)^:=default_value;
    end;
    inc(i);
  end;
end;

function SetLevel_MicroChunk(hexdata: string; const conf: array of LevelFileConfigInfo;  element: integer;  real_element: integer): string;
var
  micro_chunk_size, conf_type, byte_mask: integer;
  element_found: boolean;
  i, num_bytes: integer;
  buffer: IntArr;
  data_type, num_entities, max_num_entities: integer;
  stringa: PString;
  j, c, x, y, value: integer;
  element_array: PIntArr;
  error_conf_chunk_bytes: String;
  error_conf_chunk_token, error_element: integer;
  content: PContentArr;
  function GetHexBytes(bytes: byte): longword;
  begin
    Result:=StrToInt(HexDisplayPrefix + Copy(hexdata, 1, bytes*2));
    Delete(hexdata, 1, bytes*2);
  end;
begin
  micro_chunk_size:=0;
  conf_type:=GetHexBytes(1);
  if conf_type = 0 then
  begin
    Result:='';
    Debug('ERROR: SetLevel_MicroChunk conf_type = 0 at ' + hexdata);
    exit;
  end;
  byte_mask:=conf_type and CONF_MASK_BYTES;
  element_found:=LongBool(0);
  micro_chunk_size:= micro_chunk_size + (1);
  if byte_mask = CONF_MASK_MULTI_BYTES then
  begin
    num_bytes:=GetHexBytes(2);
    SetLength(buffer, num_bytes);
    for i:=1 to num_bytes do buffer[i-1]:=GetHexBytes(1);
    i:= 0;
    while conf[i].data_type<>-1 do
    begin
      if (conf[i].element=element)and(conf[i].conf_type=conf_type) then
      begin
        data_type:=conf[i].data_type;
        num_entities:=num_bytes div CONF_ENTITY_NUM_BYTES(data_type);
        max_num_entities:=conf[i].max_num_entities;
        if num_entities>max_num_entities then
        begin
          Debug('WARNING: truncating number of entities for element '+IntToStr(element)+' from '+IntToStr(num_entities)+' to '+IntToStr(max_num_entities));
          num_entities:= max_num_entities;
        end;
        conf[i].num_entities^:=num_entities;
        element_found:= LongBool(1);
        if data_type=TYPE_STRING then
        begin
          stringa:=PString(conf[i].value);
          SetLength(stringa^, num_entities);
          for j:=0 to Pred(num_entities) do
           stringa^[j+1]:=Chr(buffer[j])
        end
        else
        if data_type=TYPE_ELEMENT_LIST then
        begin
          element_array:=pintarr(conf[i].value);
          for j:=0 to Pred(max_num_entities) do
            element_array^[j]:= getMappedElement(CONF_ELEMENTS_ELEMENT(buffer,j));
        end
        else
        if data_type=TYPE_CONTENT_LIST then
        begin
       	  content:=PContentArr(conf[i].value);
          for c:=0 to Pred(num_entities) do
          for y:=0 to Pred(3) do
          for x:=0 to Pred(3) do
           content[c].e[x, y]:= getMappedElement(CONF_CONTENTS_ELEMENT(buffer,c,x,y));
        end
        else element_found:=LongBool(0);
        break;
      end;
      inc(i);
    end;
    if conf[i].data_type = -1 then
    begin
      Result:='';
      Exit;
    end;
    micro_chunk_size:= micro_chunk_size + (2+num_bytes);
  end
  else
  begin
    // constant size configuration data (1, 2 or 4 bytes)
    case byte_mask of
      CONF_MASK_1_BYTE: value:=GetHexBytes(1);
      CONF_MASK_2_BYTE: value:=GetHexBytes(2);
      CONF_MASK_4_BYTE: value:=GetHexBytes(4);
      else value:=0;
    end;

    i:= 0;
    while conf[i].data_type<>-1 do
    begin
      if (conf[i].element=element)and(conf[i].conf_type=conf_type) then
      begin
        data_type:=conf[i].data_type;
        if data_type=TYPE_ELEMENT then
          value:=getMappedElement(value);
        if data_type=TYPE_BOOLEAN then
          pboolean(conf[i].value)^:=Longbool(value)
        else case byte_mask of
          CONF_MASK_1_BYTE: pbyte(conf[i].value)^:=value;
          CONF_MASK_2_BYTE: pword(conf[i].value)^:=value;
          CONF_MASK_4_BYTE: pinteger(conf[i].value)^:=value;
        end;
        element_found:= LongBool(1);
        break;
      end;
      inc(i);
    end;
    micro_chunk_size:= micro_chunk_size + (CONF_VALUE_NUM_BYTES(byte_mask));
  end;
  if not element_found then
  begin
    if byte_mask=CONF_MASK_1_BYTE then
     error_conf_chunk_bytes:='CONF_VALUE_8_BIT'
    else if byte_mask=CONF_MASK_2_BYTE then error_conf_chunk_bytes:='CONF_VALUE_16_BIT'
    else if byte_mask=CONF_MASK_4_BYTE then error_conf_chunk_bytes:='CONF_VALUE_32_BIT'
    else error_conf_chunk_bytes:='CONF_VALUE_BYTES';
    error_conf_chunk_token:=conf_type and CONF_MASK_TOKEN;
    error_element:=real_element;
    Debug(Format('WARNING: cannot load micro chunk ''%d'' value for element %d',
    [conf_type, error_element]));// EL_NAME(error_element)
  end;
  result:=hexdata;
end;

function LoadLevel_MicroChunk(var filea: TByteFile; const conf: array of LevelFileConfigInfo;  element: integer;  real_element: integer): integer;
var
  micro_chunk_size, conf_type, byte_mask: integer;
  element_found: boolean;
  i, num_bytes: integer;
  buffer: IntArr;
  data_type, num_entities, max_num_entities: integer;
  stringa: PString;
  j, c, x, y, value: integer;
  element_array: PIntArr;
  error_conf_chunk_bytes: String;
  error_conf_chunk_token, error_element: integer;
  content: PContentArr;
begin
  micro_chunk_size:=0;
  conf_type:=getFile8Bit(filea);
  if conf_type = 0 then
  begin
    micro_chunk_size:=FilePos(filea);
    Result:=micro_chunk_size;
    Debug('ERROR: MicroChunk conf_type = 0 at ' + IntToStr(FilePos(filea)));
    exit;
  end;
  byte_mask:=conf_type and CONF_MASK_BYTES;
  element_found:=LongBool(0);
  micro_chunk_size:= micro_chunk_size + (1);
  if byte_mask = CONF_MASK_MULTI_BYTES then
  begin
    num_bytes:=getFile16BitBE(filea);
    SetLength(buffer, num_bytes);
    ReadBytesFromFile(filea,buffer,num_bytes);
    i:= 0;
    while conf[i].data_type<>-1 do
    begin
      if (conf[i].element=element)and(conf[i].conf_type=conf_type) then
      begin
        data_type:=conf[i].data_type;
        num_entities:=num_bytes div CONF_ENTITY_NUM_BYTES(data_type);
        max_num_entities:=conf[i].max_num_entities;
        if num_entities>max_num_entities then
        begin
          Debug('WARNING: truncating number of entities for element '+IntToStr(element)+' from '+IntToStr(num_entities)+' to '+IntToStr(max_num_entities));
          num_entities:= max_num_entities;
        end;
        
	if (num_entities = 0) and ((data_type = TYPE_ELEMENT_LIST) or
				  (data_type = TYPE_CONTENT_LIST)) then
        begin
	  { for element and content lists, zero entities are not allowed }
	  Debug('WARNING: found empty list of entities for element ' + IntToStr(element));

	  { do not set "num_entities" here to prevent reading behind buffer }

	  conf[i].num_entities^:=1;	(* at least one is required *)
        end
	else conf[i].num_entities^:=num_entities;
        
        element_found:= LongBool(1);
        if data_type=TYPE_STRING then
        begin
          stringa:=PString(conf[i].value);
          SetLength(stringa^, num_entities);
          for j:=0 to Pred(num_entities) do
           stringa^[j+1]:=Chr(buffer[j])
        end
        else
        if data_type=TYPE_ELEMENT_LIST then
        begin
          element_array:=pintarr(conf[i].value);
          for j:=0 to Pred(max_num_entities) do
            element_array^[j]:= getMappedElement(CONF_ELEMENTS_ELEMENT(buffer,j));
        end
        else
        if data_type=TYPE_CONTENT_LIST then
        begin
       	  content:=PContentArr(conf[i].value);
          for c:=0 to Pred(num_entities) do
          for y:=0 to Pred(3) do
          for x:=0 to Pred(3) do
           content[c].e[x, y]:= getMappedElement(CONF_CONTENTS_ELEMENT(buffer,c,x,y));
        end
        else element_found:=LongBool(0);
        break;
      end;
      inc(i);
    end;
    micro_chunk_size:= micro_chunk_size + (2+num_bytes);
  end
  else
  begin
    // constant size configuration data (1, 2 or 4 bytes)
    case byte_mask of
      CONF_MASK_1_BYTE: value:=getFile8Bit(filea);
      CONF_MASK_2_BYTE: value:=getFile16BitBE(filea);
      CONF_MASK_4_BYTE: value:=getFile32BitBE(filea);
      else value:=0;
    end;

    i:= 0;
    while conf[i].data_type<>-1 do
    begin
      if (conf[i].element=element)and(conf[i].conf_type=conf_type) then
      begin
        data_type:=conf[i].data_type;
        if data_type=TYPE_ELEMENT then
          value:=getMappedElement(value);
        if data_type=TYPE_BOOLEAN then
          pboolean(conf[i].value)^:=Longbool(value)
        else case byte_mask of
          CONF_MASK_1_BYTE: pbyte(conf[i].value)^:=value;
          CONF_MASK_2_BYTE: pword(conf[i].value)^:=value;
          CONF_MASK_4_BYTE: pinteger(conf[i].value)^:=value;
//          else value:=0;
        end;
        element_found:= LongBool(1);
        break;
      end;
      inc(i);
    end;
    micro_chunk_size:= micro_chunk_size + (CONF_VALUE_NUM_BYTES(byte_mask));
  end;
  if not element_found then
  begin
    if byte_mask=CONF_MASK_1_BYTE then
     error_conf_chunk_bytes:='CONF_VALUE_8_BIT'
    else if byte_mask=CONF_MASK_2_BYTE then error_conf_chunk_bytes:='CONF_VALUE_16_BIT'
    else if byte_mask=CONF_MASK_4_BYTE then error_conf_chunk_bytes:='CONF_VALUE_32_BIT'
    else error_conf_chunk_bytes:='CONF_VALUE_BYTES';
    error_conf_chunk_token:=conf_type and CONF_MASK_TOKEN;
    error_element:=real_element;
    Debug(Format('WARNING: cannot load micro chunk ''%d'' value for element %d',
    [conf_type, error_element]));// EL_NAME(error_element)
  end;
  result:=micro_chunk_size;
end;

function SaveLevel_MicroChunk(var filea: TByteFile; entry: pLevelFileConfigInfo; write_element: boolean): integer;
var
  save_type, data_type, conf_type, byte_mask, element, default_value, num_bytes: integer;
  modified: boolean;
  value_ptr: pointer;
  default_string: pstring;
  stringa: pstring;
  element_array: PIntArr;
  value, string_length, num_elements, num_contents, i, x, y: integer;
  content: PContentArr;
begin
  save_type:=entry.save_type;
  data_type:=entry.data_type;
  conf_type:=entry.conf_type;
  byte_mask:=conf_type and CONF_MASK_BYTES;
  element:=entry.element;
  default_value:=entry.default_value;
  num_bytes:=0;
  modified:=LongBool(0);
  if byte_mask<>CONF_MASK_MULTI_BYTES then
  begin
    value_ptr:=entry.value;
    if data_type=TYPE_BOOLEAN then value:=integer(pboolean(value_ptr)^)
    else value:=pinteger(value_ptr)^;
    (* check if any settings have been modified before saving them *)
    if value<>default_value then
    modified:= LongBool(1);
    (* do not save if explicitly told or if unmodified default settings *)
    if (save_type = SAVE_CONF_NEVER) or ((save_type = SAVE_CONF_WHEN_CHANGED) and (not modified)) then
    begin
      result:= 0;
      exit;
    end;
    if write_element then
      num_bytes:= num_bytes + (putFile16BitBE(filea,element));

    num_bytes:= num_bytes + (putFile8Bit(filea,conf_type));
    if (byte_mask=CONF_MASK_1_BYTE) then num_bytes:=num_bytes + putFile8Bit(filea,value)
    else if byte_mask=CONF_MASK_2_BYTE then num_bytes:=num_bytes + putFile16BitBE(filea,value)
    else if byte_mask=CONF_MASK_4_BYTE then num_bytes:=num_bytes + putFile32BitBE(filea,value);
  end
  else
  if data_type=TYPE_STRING then
  begin
    default_string:=PString(entry.default_string);
    stringa:=PString(entry.value);
    string_length:=length(stringa^);
    if not (stringa^ = default_string^) then
     modified:= LongBool(1);
    (* do not save if explicitly told or if unmodified default settings *)
    if (save_type=SAVE_CONF_NEVER) or ((save_type=SAVE_CONF_WHEN_CHANGED)and (not modified)) then
    begin
      result:=0;
      exit;
    end;
    if write_element then
      num_bytes:= num_bytes + (putFile16BitBE(filea,element));

    num_bytes:= num_bytes + (putFile8Bit(filea,conf_type));
    num_bytes:= num_bytes + (putFile16BitBE(filea,string_length));
    for i:=0 to Pred(string_length) do
     num_bytes:= num_bytes + (putFile8Bit(filea,ord(stringa^[i+1])));
  end
  else
  if data_type=TYPE_ELEMENT_LIST then
  begin
    element_array:=PIntArr(entry.value);
    num_elements:=pinteger(entry.num_entities)^;
    for i:=0 to Pred(num_elements) do
    if element_array[i] <> default_value then
     modified:=LongBool(1);
    (* do not save if explicitly told or if unmodified default settings *)
    if (save_type = SAVE_CONF_NEVER) or ((save_type = SAVE_CONF_WHEN_CHANGED) and (not modified))
    then
    begin
      result:=0;
      exit;
    end;
    if write_element then
      num_bytes:= num_bytes + (putFile16BitBE(filea,element));

    num_bytes:= num_bytes + (putFile8Bit(filea,conf_type));
    num_bytes:= num_bytes + (putFile16BitBE(filea,num_elements*CONF_ELEMENT_NUM_BYTES));
    for i:=0 to Pred(num_elements) do
      num_bytes:= num_bytes + (putFile16BitBE(filea,element_array[i]));
  end
  else if data_type=TYPE_CONTENT_LIST then
  begin
    content:=PContentArr(entry.value);
    num_contents:=pinteger(entry.num_entities)^;
    for i:=0 to Pred(num_contents) do
    for y:=0 to Pred(3) do
    for x:=0 to Pred(3) do
    if content[i].e[x, y] <> default_value then
      modified:= LongBool(1);
    (* do not save if explicitly told or if unmodified default settings *)

    if (save_type=SAVE_CONF_NEVER) or ((save_type=SAVE_CONF_WHEN_CHANGED)and (not modified))
    then
    begin
      result:=0;
      exit;
    end;
    if write_element then
      num_bytes:= num_bytes + (putFile16BitBE(filea,element));

    num_bytes:= num_bytes + (putFile8Bit(filea,conf_type));
    num_bytes:= num_bytes + (putFile16BitBE(filea,num_contents*CONF_CONTENT_NUM_BYTES));
    for i:=0 to Pred(num_contents) do
    for y:=0 to Pred(3) do
    for x:=0 to Pred(3) do
     num_bytes:= num_bytes + (putFile16BitBE(filea,content[i].e[x, y]));
  end;
  result:= num_bytes;
  exit;
end;

end.
